package de.juplo.kafka.payment.transfer.adapter;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import de.juplo.kafka.payment.transfer.domain.Transfer;
import de.juplo.kafka.payment.transfer.ports.CreateTransferUseCase;
import de.juplo.kafka.payment.transfer.ports.GetTransferUseCase;
import de.juplo.kafka.payment.transfer.ports.HandleStateChangeUseCase;
import de.juplo.kafka.payment.transfer.ports.TransferRepository;
import lombok.extern.slf4j.Slf4j;
import org.apache.kafka.clients.admin.AdminClient;
import org.apache.kafka.clients.admin.MemberDescription;
import org.apache.kafka.clients.consumer.ConsumerRebalanceListener;
import org.apache.kafka.clients.consumer.ConsumerRecord;
import org.apache.kafka.clients.consumer.ConsumerRecords;
import org.apache.kafka.clients.consumer.KafkaConsumer;
import org.apache.kafka.common.TopicPartition;
import org.apache.kafka.common.errors.WakeupException;
import org.springframework.context.event.ContextRefreshedEvent;
import org.springframework.context.event.EventListener;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.ResponseBody;

import java.time.Clock;
import java.time.Duration;
import java.time.Instant;
import java.util.*;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.function.Consumer;


@RequestMapping("/consumer")
@ResponseBody
@Slf4j
public class TransferConsumer implements Runnable, ConsumerRebalanceListener
{
  private final String topic;
  private final int numPartitions;
  private final KafkaConsumer<String, String> consumer;
  private final AdminClient adminClient;
  private final TransferRepository repository;
  private final ObjectMapper mapper;
  private final ConsumerUseCases restoreUseCases;

  private boolean running = false;
  private boolean shutdown = false;
  private Future<?> future = null;

  private final String groupId;
  private final String groupInstanceId;
  private final Map<String, String> instanceIdUriMapping;
  private final String[] instanceIdByPartition;


  private Clock clock;
  private int stateStoreInterval;

  private final Consumer<ConsumerRecord<String, String>> productionRecordHandler;
  private final Consumer<ConsumerRecord<String, String>> recordHandlers[];

  private volatile boolean partitionOwnershipUnknown = true;


  public TransferConsumer(
      String topic,
      int numPartitions,
      Map<String, String> instanceIdUriMapping,
      KafkaConsumer<String, String> consumer,
      AdminClient adminClient,
      TransferRepository repository,
      Clock clock,
      int stateStoreInterval,
      ObjectMapper mapper,
      ConsumerUseCases productionUseCases,
      ConsumerUseCases restoreUseCases)
  {
    this.topic = topic;
    this.numPartitions = numPartitions;
    this.groupId = consumer.groupMetadata().groupId();
    this.groupInstanceId = consumer.groupMetadata().groupInstanceId().get();
    this.instanceIdByPartition = new String[numPartitions];
    this.instanceIdUriMapping = new HashMap<>(instanceIdUriMapping.size());
    for (String instanceId : instanceIdUriMapping.keySet())
    {
      // Requests are not redirected for the instance itself
      String uri = instanceId.equals(groupInstanceId)
          ? null
          : instanceIdUriMapping.get(instanceId);
      this.instanceIdUriMapping.put(instanceId, uri);
    }
    this.consumer = consumer;
    this.adminClient = adminClient;
    this.repository = repository;
    this.clock = clock;
    this.stateStoreInterval = stateStoreInterval;
    this.mapper = mapper;
    this.restoreUseCases = restoreUseCases;

    productionRecordHandler = (record) -> handleRecord(record, productionUseCases);
    this.recordHandlers = new Consumer[numPartitions];
  }


  @Override
  public void run()
  {
    Instant stateStored = clock.instant();

    while (running)
    {
      try
      {
        ConsumerRecords<String, String> records = consumer.poll(Duration.ofSeconds(1));
        if (records.count() == 0)
          continue;

        log.debug("polled {} records", records.count());
        records.forEach(record -> recordHandlers[record.partition()].accept(record));

        Instant now = clock.instant();
        if (
            stateStoreInterval > 0 &&
            Duration.between(stateStored, now).getSeconds() >= stateStoreInterval)
        {
          Map<Integer, Long> offsets = new HashMap<>();

          for (TopicPartition topicPartition : consumer.assignment())
          {
            Integer partition = topicPartition.partition();
            Long offset = consumer.position(topicPartition);
            log.info("storing state locally for {}/{}: {}", topic, partition, offset);
            offsets.put(partition, offset);
          }

          repository.storeState(offsets);
          stateStored = now;
        }
      }
      catch (WakeupException e)
      {
        log.info("cleanly interrupted while polling");
      }
    }

    log.info("polling stopped");
  }

  private void handleRecord(ConsumerRecord<String, String> record, ConsumerUseCases useCases)
  {
    try
    {
      byte eventType = record.headers().lastHeader(EventType.HEADER).value()[0];

      switch (eventType)
      {
        case EventType.NEW_TRANSFER:

          NewTransferEvent newTransferEvent =
              mapper.readValue(record.value(), NewTransferEvent.class);
          useCases
              .create(
                  newTransferEvent.getId(),
                  newTransferEvent.getPayer(),
                  newTransferEvent.getPayee(),
                  newTransferEvent.getAmount());
          break;

        case EventType.TRANSFER_STATE_CHANGED:

          TransferStateChangedEvent stateChangedEvent =
              mapper.readValue(record.value(), TransferStateChangedEvent.class);
          useCases.handleStateChange(stateChangedEvent);
          break;
      }
    }
    catch (JsonProcessingException e)
    {
      log.error(
          "ignoring invalid json in message #{} on {}/{}: {}",
          record.offset(),
          record.topic(),
          record.partition(),
          record.value());
    }
    catch (IllegalArgumentException e)
    {
      log.error(
          "ignoring invalid message #{} on {}/{}: {}, message={}",
          record.offset(),
          record.topic(),
          record.partition(),
          e.getMessage(),
          record.value());
    }
  }


  /**
   * Identifies the URI, at which the Group-Instance can be reached,
   * that holds the state for a specific {@link Transfer}.
   *
   * The {@link Transfer#getId() ID} of the {@link Transfer} is named
   * {@code key} here and of type {@code String}, because this example
   * project stores the key as a String in Kafka to simplify the listing
   * and manual manipulation of the according topic.
   *
   * @param key A {@code String}, that represents the {@link Transfer#getId() ID} of a {@link Transfer}.
   * @return An {@link Optional}, that holds the URI at which the Group-Instance
   * can be reached, that holds the state for the {@link Transfer}, that
   * is identified by the key (if present), or is empty, if the {@link Transfer}
   * would be handled by the local instance.
   */
  public Optional<String> uriForKey(String key)
  {
    synchronized (this)
    {
      while (partitionOwnershipUnknown)
      {
        try { wait(); } catch (InterruptedException e) {}
      }

      int partition = TransferPartitioner.computeHashForKey(key, numPartitions);
      return
          Optional
              .ofNullable(instanceIdByPartition[partition])
              .map(id -> instanceIdUriMapping.get(id));
    }
  }

  @EventListener
  public synchronized void onApplicationEvent(ContextRefreshedEvent event)
  {
    // "Needed", because this method is called synchronously during the
    // initialization pahse of Spring. If the subscription happens
    // in the same thread, it would block the completion of the initialization.
    // Hence, the app would not react to any signal (CTRL-C, for example) except
    // a KILL until the restoring is finished.
    future = CompletableFuture.runAsync(() -> start());
    log.info("start of application completed");
  }


  @Override
  public void onPartitionsRevoked(Collection<TopicPartition> partitions)
  {
    partitionOwnershipUnknown = true;
    log.info("partitions revoked: {}", partitions);
    for (TopicPartition topicPartition : partitions)
    {
      int partition = topicPartition.partition();
      long offset = consumer.position(topicPartition);
      log.info("deactivating partition {}, offset: {}", partition, offset);
      repository.deactivatePartition(partition, offset);
    }
  }

  @Override
  public void onPartitionsAssigned(Collection<TopicPartition> partitions)
  {
    log.info("partitions assigned: {}", partitions);
    fetchAssignmentsAsync();
    if (partitions.size() > 0)
    {
      for (Map.Entry<TopicPartition, Long> entry : consumer.endOffsets(partitions).entrySet())
      {
        TopicPartition topicPartition = entry.getKey();
        Integer partition = topicPartition.partition();
        long offset = repository.activatePartition(partition);
        log.info("activated partition {}, seeking to offset {}", partition, offset);
        consumer.seek(topicPartition, offset);
        Long endOffset = entry.getValue();
        if (offset < endOffset)
        {
          log.info("--> starting restore of partition {}: {} -> {}", partition, offset, endOffset);
          recordHandlers[partition] = new RestoreRecordHandler(endOffset);
        }
        else
        {
          log.info("--> partition {} is up-to-date, offset: {}", partition, offset);
          recordHandlers[partition] = productionRecordHandler;
        }
      }
    }
  }

  private void fetchAssignmentsAsync()
  {
    adminClient
        .describeConsumerGroups(List.of(groupId))
        .describedGroups()
        .get(groupId)
        .whenComplete((descriptions, e) ->
        {
          if (e != null)
          {
            log.error("could not fetch group data: {}", e.getMessage());
          }
          else
          {
            synchronized (this)
            {
              for (MemberDescription description : descriptions.members())
              {
                description
                    .assignment()
                    .topicPartitions()
                    .forEach(tp -> instanceIdByPartition[tp.partition()] = description.groupInstanceId().get());
              }
              partitionOwnershipUnknown = false;
              notifyAll();
            }
          }
        });
  }

  @Override
  public void onPartitionsLost(Collection<TopicPartition> partitions)
  {
    partitionOwnershipUnknown = true;
    log.info("partiotions lost: {}", partitions);
  }


  class RestoreRecordHandler implements Consumer<ConsumerRecord<String, String>>
  {
    final long seen;


    RestoreRecordHandler(Long endOffset)
    {
      this.seen = endOffset - 1;
    }


    @Override
    public void accept(ConsumerRecord<String, String> record)
    {
      if (seen < record.offset())
      {
        int partition = record.partition();
        log.info(
            "--> restore of partition {} completed: needed={}, seen={}!",
            partition,
            seen,
            record.offset());
        recordHandlers[partition] = productionRecordHandler;
        productionRecordHandler.accept(record);
      }
      else
      {
        handleRecord(record, restoreUseCases);
        if (seen == record.offset())
        {
          int partition = record.partition();
          log.info( "--> restore of partition {} completed!", partition);
          recordHandlers[partition] = productionRecordHandler;
        }
      }
    }
  }


  @PostMapping("start")
  public synchronized String start()
  {
    if (running)
    {
      log.info("consumer already running!");
      return "Already running!";
    }

    int foundNumPartitions = consumer.partitionsFor(topic).size();
    if (foundNumPartitions != numPartitions)
    {
      log.error(
          "unexpected number of partitions for topic {}: expected={}, found={}",
          topic,
          numPartitions,
          foundNumPartitions
          );
      return "Wrong number of partitions for topic " + topic + ": " + foundNumPartitions;
    }

    consumer.subscribe(List.of(topic), this);

    running = true;
    future = CompletableFuture.runAsync(this);

    log.info("consumer started");
    return "Started";
  }

  @PostMapping("stop")
  public synchronized String stop()
  {
    if (!running)
    {
      log.info("consumer not running!");
      return "Not running";
    }

    running = false;

    if (!future.isDone())
      consumer.wakeup();

    log.info("waiting for the consumer...");
    try
    {
      future.get();
    }
    catch (InterruptedException|ExecutionException e)
    {
      log.error("Exception while joining polling task!", e);
      return e.getMessage();
    }
    finally
    {
      future = null;
      consumer.unsubscribe();
    }

    log.info("consumer stopped");
    return "Stopped";
  }

  public synchronized void shutdown()
  {
    log.info("shutdown initiated!");
    shutdown = true;
    stop();
    log.info("closing consumer");
    consumer.close();
  }



  public interface ConsumerUseCases
      extends
        GetTransferUseCase,
        CreateTransferUseCase,
        HandleStateChangeUseCase {};
}
